"""
Utilities to generate embedded C code from OSQP sources
"""
# Compatibility with Python 2
from __future__ import print_function
from builtins import range

# Path of osqp module
import os.path
import osqp
files_to_generate_path = os.path.join(osqp.__path__[0],
                                      'codegen', 'files_to_generate')

# Timestamp
import datetime


def write_vec(f, vec, name, vec_type):
    """
    Write vector to file
    """
    if len(vec) > 0:

        f.write('%s %s[%d] = {\n' % (vec_type, name, len(vec)))

        # Write vector elements
        for i in range(len(vec)):
            if vec_type == 'c_float':
                f.write('(c_float)%.20f,\n' % vec[i])
            else:
                f.write('%i,\n' % vec[i])

        f.write('};\n')


def write_vec_extern(f, vec, name, vec_type):
    """
    Write vector prototype to file
    """
    if len(vec) > 0:
        f.write("extern %s %s[%d];\n" % (vec_type, name, len(vec)))


def write_mat(f, mat, name):
    """
    Write scipy sparse matrix in CSC form to file
    """
    write_vec(f, mat['p'], name + '_p', 'c_int')
    if len(mat['x']) > 0:
        write_vec(f, mat['i'], name + '_i', 'c_int')
        write_vec(f, mat['x'], name + '_x', 'c_float')

    f.write("csc %s = {" % name)
    f.write("%d, " % mat['nzmax'])
    f.write("%d, " % mat['m'])
    f.write("%d, " % mat['n'])
    f.write("%s_p, " % name)
    if len(mat['x']) > 0:
        f.write("%s_i, " % name)
        f.write("%s_x, " % name)
    else:
        f.write("0, 0, ")
    f.write("%d};\n" % mat['nz'])


def write_mat_extern(f, mat, name):
    """
    Write matrix prototype to file
    """
    f.write("extern csc %s;\n" % name)


def write_data_src(f, data):
    """
    Write data structure to file
    """
    f.write("// Define data structure\n")

    # Define matrix P
    write_mat(f, data['P'], 'Pdata')

    # Define matrix A
    write_mat(f, data['A'], 'Adata')

    # Define other data vectors
    write_vec(f, data['q'], 'qdata', 'c_float')
    write_vec(f, data['l'], 'ldata', 'c_float')
    write_vec(f, data['u'], 'udata', 'c_float')

    # Define data structure
    f.write("OSQPData data = {")
    f.write("%d, " % data['n'])
    f.write("%d, " % data['m'])
    f.write("&Pdata, &Adata, qdata, ldata, udata")
    f.write("};\n\n")


def write_data_inc(f, data):
    """
    Write data structure prototypes to file
    """
    f.write("// Data structure prototypes\n")

    # Define matrix P
    write_mat_extern(f, data['P'], 'Pdata')

    # Define matrix A
    write_mat_extern(f, data['A'], 'Adata')

    # Define other data vectors
    write_vec_extern(f, data['q'], 'qdata', 'c_float')
    write_vec_extern(f, data['l'], 'ldata', 'c_float')
    write_vec_extern(f, data['u'], 'udata', 'c_float')

    # Define data structure
    f.write("extern OSQPData data;\n\n")


def write_settings_src(f, settings, embedded_flag):
    """
    Write settings structure to file
    """
    f.write("// Define settings structure\n")
    f.write("OSQPSettings settings = {")
    f.write("(c_float)%.20f, " % settings['rho'])
    f.write("(c_float)%.20f, " % settings['sigma'])
    f.write("%d, " % settings['scaling'])

    if embedded_flag != 1:
        f.write("%d, " % settings['adaptive_rho'])
        f.write("%d, " % settings['adaptive_rho_interval'])
        f.write("(c_float)%.20f, " % settings['adaptive_rho_tolerance'])

    f.write("%d, " % settings['max_iter'])
    f.write("(c_float)%.20f, " % settings['eps_abs'])
    f.write("(c_float)%.20f, " % settings['eps_rel'])
    f.write("(c_float)%.20f, " % settings['eps_prim_inf'])
    f.write("(c_float)%.20f, " % settings['eps_dual_inf'])
    f.write("(c_float)%.20f, " % settings['alpha'])
    f.write("(enum linsys_solver_type) LINSYS_SOLVER, ")

    f.write("%d, " % settings['scaled_termination'])
    f.write("%d, " % settings['check_termination'])
    f.write("%d, " % settings['warm_start'])

    f.write("};\n\n")


def write_settings_inc(f, settings, embedded_flag):
    """
    Write prototype for settings structure to file
    """
    f.write("// Settings structure prototype\n")
    f.write("extern OSQPSettings settings;\n\n")


def write_scaling_src(f, scaling):
    """
    Write scaling structure to file
    """
    f.write("// Define scaling structure\n")
    if scaling is not None:
        write_vec(f, scaling['D'],    'Dscaling',    'c_float')
        write_vec(f, scaling['Dinv'], 'Dinvscaling', 'c_float')
        write_vec(f, scaling['E'],    'Escaling',    'c_float')
        write_vec(f, scaling['Einv'], 'Einvscaling', 'c_float')
        f.write("OSQPScaling scaling = {")
        f.write("(c_float)%.20f, " % scaling['c'])
        f.write("Dscaling, Escaling, ")
        f.write("(c_float)%.20f, " % scaling['cinv'])
        f.write("Dinvscaling, Einvscaling};\n\n")
    else:
        f.write("OSQPScaling scaling;\n\n")


def write_scaling_inc(f, scaling):
    """
    Write prototypes for the scaling structure to file
    """
    f.write("// Scaling structure prototypes\n")

    if scaling is not None:
        write_vec_extern(f, scaling['D'],    'Dscaling',    'c_float')
        write_vec_extern(f, scaling['Dinv'], 'Dinvscaling', 'c_float')
        write_vec_extern(f, scaling['E'],    'Escaling',    'c_float')
        write_vec_extern(f, scaling['Einv'], 'Einvscaling', 'c_float')

    f.write("extern OSQPScaling scaling;\n\n")


def write_linsys_solver_src(f, linsys_solver, embedded_flag):
    """
    Write linsys_solver structure to file
    """

    f.write("// Define linsys_solver structure\n")
    write_mat(f, linsys_solver['L'],            'linsys_solver_L')
    write_vec(f, linsys_solver['Dinv'],         'linsys_solver_Dinv',           'c_float')
    write_vec(f, linsys_solver['P'],            'linsys_solver_P',              'c_int')
    f.write("c_float linsys_solver_bp[%d];\n"  % (len(linsys_solver['bp'])))
    f.write("c_float linsys_solver_sol[%d];\n" % (len(linsys_solver['sol'])))
    write_vec(f, linsys_solver['rho_inv_vec'],  'linsys_solver_rho_inv_vec',    'c_float')

    if embedded_flag != 1:
        write_vec(f, linsys_solver['Pdiag_idx'], 'linsys_solver_Pdiag_idx', 'c_int')
        write_mat(f, linsys_solver['KKT'],       'linsys_solver_KKT')
        write_vec(f, linsys_solver['PtoKKT'],    'linsys_solver_PtoKKT',    'c_int')
        write_vec(f, linsys_solver['AtoKKT'],    'linsys_solver_AtoKKT',    'c_int')
        write_vec(f, linsys_solver['rhotoKKT'],  'linsys_solver_rhotoKKT',  'c_int')
        write_vec(f, linsys_solver['D'],         'linsys_solver_D',         'QDLDL_float')
        write_vec(f, linsys_solver['etree'],     'linsys_solver_etree',     'QDLDL_int')
        write_vec(f, linsys_solver['Lnz'],       'linsys_solver_Lnz',       'QDLDL_int')
        f.write("QDLDL_int   linsys_solver_iwork[%d];\n" % len(linsys_solver['iwork']))
        f.write("QDLDL_bool  linsys_solver_bwork[%d];\n" % len(linsys_solver['bwork']))
        f.write("QDLDL_float linsys_solver_fwork[%d];\n" % len(linsys_solver['fwork']))

    f.write("qdldl_solver linsys_solver = ")
    f.write("{QDLDL_SOLVER, &solve_linsys_qdldl, ")

    if embedded_flag != 1:
        f.write("&update_linsys_solver_matrices_qdldl, &update_linsys_solver_rho_vec_qdldl, ")

    f.write("&linsys_solver_L, linsys_solver_Dinv, linsys_solver_P, linsys_solver_bp, linsys_solver_sol, linsys_solver_rho_inv_vec, ")
    f.write("(c_float)%.20f, " % linsys_solver['sigma'])
    f.write("%d, " % linsys_solver['n'])
    f.write("%d, " % linsys_solver['m'])
    
    if embedded_flag != 1:
        if len(linsys_solver['Pdiag_idx']) > 0:
            linsys_solver_Pdiag_idx_string = 'linsys_solver_Pdiag_idx'
            linsys_solver_PtoKKT_string = 'linsys_solver_PtoKKT'
        else:
            linsys_solver_Pdiag_idx_string = '0'
            linsys_solver_PtoKKT_string = '0'
        if len(linsys_solver['AtoKKT']) > 0:
            linsys_solver_AtoKKT_string = 'linsys_solver_AtoKKT'
        else:
            linsys_solver_AtoKKT_string = '0'
        f.write("%s, " % linsys_solver_Pdiag_idx_string)
        f.write("%d, " % linsys_solver['Pdiag_n'])
        f.write("&linsys_solver_KKT, %s, %s, linsys_solver_rhotoKKT, "
                % (linsys_solver_PtoKKT_string, linsys_solver_AtoKKT_string) +
                "linsys_solver_D, linsys_solver_etree, linsys_solver_Lnz, " +
                "linsys_solver_iwork, linsys_solver_bwork, linsys_solver_fwork, ")
    
    f.write("};\n\n")


def write_linsys_solver_inc(f, linsys_solver, embedded_flag):
    """
    Write prototypes for linsys_solver structure to file
    """
    f.write("// Prototypes for linsys_solver structure\n")
    write_mat_extern(f, linsys_solver['L'],    'linsys_solver_L')
    write_vec_extern(f, linsys_solver['Dinv'], 'linsys_solver_Dinv', 'c_float')
    write_vec_extern(f, linsys_solver['P'],    'linsys_solver_P',    'c_int')
    f.write("extern c_float linsys_solver_bp[%d];\n"  % len(linsys_solver['bp']))
    f.write("extern c_float linsys_solver_sol[%d];\n" % len(linsys_solver['sol']))
    write_vec_extern(f, linsys_solver['rho_inv_vec'], 'linsys_solver_rho_inv_vec', 'c_float')

    if embedded_flag != 1:
        write_vec_extern(f, linsys_solver['Pdiag_idx'], 'linsys_solver_Pdiag_idx', 'c_int')
        write_mat_extern(f, linsys_solver['KKT'],       'linsys_solver_KKT')
        write_vec_extern(f, linsys_solver['PtoKKT'],    'linsys_solver_PtoKKT',    'c_int')
        write_vec_extern(f, linsys_solver['AtoKKT'],    'linsys_solver_AtoKKT',    'c_int')
        write_vec_extern(f, linsys_solver['rhotoKKT'],  'linsys_solver_rhotoKKT',  'c_int')
        write_vec_extern(f, linsys_solver['D'],         'linsys_solver_D',         'QDLDL_float')
        write_vec_extern(f, linsys_solver['etree'],     'linsys_solver_etree',     'QDLDL_int')
        write_vec_extern(f, linsys_solver['Lnz'],       'linsys_solver_Lnz',       'QDLDL_int')
        f.write("extern QDLDL_int   linsys_solver_iwork[%d];\n" % len(linsys_solver['iwork']))
        f.write("extern QDLDL_bool  linsys_solver_bwork[%d];\n" % len(linsys_solver['bwork']))
        f.write("extern QDLDL_float linsys_solver_fwork[%d];\n" % len(linsys_solver['fwork']))

    f.write("extern qdldl_solver linsys_solver;\n\n")


def write_solution_src(f, data):
    """
    Preallocate solution vectors
    """
    f.write("// Define solution\n")
    f.write("c_float xsolution[%d];\n" % data['n'])
    f.write("c_float ysolution[%d];\n\n" % data['m'])
    f.write("OSQPSolution solution = {xsolution, ysolution};\n\n")


def write_solution_inc(f, data):
    """
    Prototypes for solution vectors
    """
    f.write("// Prototypes for solution\n")
    f.write("extern c_float xsolution[%d];\n" % data['n'])
    f.write("extern c_float ysolution[%d];\n\n" % data['m'])
    f.write("extern OSQPSolution solution;\n\n")


def write_info_src(f):
    """
    Preallocate info structure
    """
    f.write("// Define info\n")
    f.write('OSQPInfo info = {0, "Unsolved", OSQP_UNSOLVED, 0.0, 0.0, 0.0};\n\n')


def write_info_inc(f):
    """
    Prototype for info structure
    """
    f.write("// Prototype for info structure\n")
    f.write("extern OSQPInfo info;\n\n")


def write_workspace_src(f, n, m, rho_vectors, embedded_flag):
    """
    Preallocate workspace structure and populate rho vectors
    """

    f.write("// Define workspace\n")

    write_vec(f, rho_vectors['rho_vec'],     'work_rho_vec',     'c_float')
    write_vec(f, rho_vectors['rho_inv_vec'], 'work_rho_inv_vec', 'c_float')
    if embedded_flag != 1:
        write_vec(f, rho_vectors['constr_type'], 'work_constr_type', 'c_int')

    f.write("c_float work_x[%d];\n" % n)
    f.write("c_float work_y[%d];\n" % m)
    f.write("c_float work_z[%d];\n" % m)
    f.write("c_float work_xz_tilde[%d];\n" % (m + n))
    f.write("c_float work_x_prev[%d];\n" % n)
    f.write("c_float work_z_prev[%d];\n" % m)
    f.write("c_float work_Ax[%d];\n" % m)
    f.write("c_float work_Px[%d];\n" % n)
    f.write("c_float work_Aty[%d];\n" % n)
    f.write("c_float work_delta_y[%d];\n" % m)
    f.write("c_float work_Atdelta_y[%d];\n" % n)
    f.write("c_float work_delta_x[%d];\n" % n)
    f.write("c_float work_Pdelta_x[%d];\n" % n)
    f.write("c_float work_Adelta_x[%d];\n" % m)
    f.write("c_float work_D_temp[%d];\n" % n)
    f.write("c_float work_D_temp_A[%d];\n" % n)
    f.write("c_float work_E_temp[%d];\n\n" % m)

    f.write("OSQPWorkspace workspace = {\n")
    f.write("&data, (LinSysSolver *)&linsys_solver,\n")
    f.write("work_rho_vec, work_rho_inv_vec,\n")
    if embedded_flag != 1:
        f.write("work_constr_type,\n")

    f.write("work_x, work_y, work_z, work_xz_tilde,\n")
    f.write("work_x_prev, work_z_prev,\n")
    f.write("work_Ax, work_Px, work_Aty,\n")
    f.write("work_delta_y, work_Atdelta_y,\n")
    f.write("work_delta_x, work_Pdelta_x, work_Adelta_x,\n")
    f.write("work_D_temp, work_D_temp_A, work_E_temp,\n")
    f.write("&settings, &scaling, &solution, &info};\n\n")


def write_workspace_inc(f, n, m, rho_vectors, embedded_flag):
    """
    Prototypes for the workspace structure and rho_vectors
    """
    f.write("// Prototypes for the workspace\n")
    write_vec_extern(f, rho_vectors['rho_vec'],     'work_rho_vec',     'c_float')
    write_vec_extern(f, rho_vectors['rho_inv_vec'], 'work_rho_inv_vec', 'c_float')
    if embedded_flag != 1:
        write_vec_extern(f, rho_vectors['constr_type'], 'work_constr_type', 'c_int')

    f.write("extern c_float work_x[%d];\n" % n)
    f.write("extern c_float work_y[%d];\n" % m)
    f.write("extern c_float work_z[%d];\n" % m)
    f.write("extern c_float work_xz_tilde[%d];\n" % (m + n))
    f.write("extern c_float work_x_prev[%d];\n" % n)
    f.write("extern c_float work_z_prev[%d];\n" % m)
    f.write("extern c_float work_Ax[%d];\n" % m)
    f.write("extern c_float work_Px[%d];\n" % n)
    f.write("extern c_float work_Aty[%d];\n" % n)
    f.write("extern c_float work_delta_y[%d];\n" % m)
    f.write("extern c_float work_Atdelta_y[%d];\n" % n)
    f.write("extern c_float work_delta_x[%d];\n" % n)
    f.write("extern c_float work_Pdelta_x[%d];\n" % n)
    f.write("extern c_float work_Adelta_x[%d];\n" % m)
    f.write("extern c_float work_D_temp[%d];\n" % n)
    f.write("extern c_float work_D_temp_A[%d];\n" % n)
    f.write("extern c_float work_E_temp[%d];\n\n" % m)

    f.write("extern OSQPWorkspace workspace;\n\n")


def render_workspace(variables, hfname, cfname):
    """
    Print workspace dimensions
    """

    rho_vectors = variables['rho_vectors']
    data = variables['data']
    linsys_solver = variables['linsys_solver']
    scaling = variables['scaling']
    settings = variables['settings']
    embedded_flag = variables['embedded_flag']
    n = data['n']
    m = data['m']

    # Open output file
    incFile = open(hfname, 'w')
    srcFile = open(cfname, 'w')

    # Add an include-guard statement
    fname = os.path.splitext(os.path.basename(hfname))[0]
    incGuard = fname.upper() + "_H"
    incFile.write("#ifndef %s\n" % incGuard)
    incFile.write("#define %s\n\n" % incGuard)

    # Print comment headers containing the generation time into the files
    now = datetime.datetime.now()
    daystr = now.strftime("%B %d, %Y")
    timestr = now.strftime("%H:%M:%S")
    incFile.write("/*\n")
    incFile.write(" * This file was autogenerated by OSQP-Python on %s at %s.\n" % (daystr, timestr))
    incFile.write(" * \n")
    incFile.write(" * This file contains the prototypes for all the workspace variables needed\n")
    incFile.write(" * by OSQP. The actual data is contained inside workspace.c.\n")
    incFile.write(" */\n\n")

    srcFile.write("/*\n")
    srcFile.write(" * This file was autogenerated by OSQP-Python on %s at %s.\n" % (daystr, timestr))
    srcFile.write(" * \n")
    srcFile.write(" * This file contains the workspace variables needed by OSQP.\n")
    srcFile.write(" */\n\n")

    # Include types, constants and linsys_solver header
    incFile.write("#include \"types.h\"\n")
    incFile.write("#include \"qdldl_interface.h\"\n\n")

    srcFile.write("#include \"types.h\"\n")
    srcFile.write("#include \"qdldl_interface.h\"\n\n")

    # Write data structure
    write_data_src(srcFile, data)
    write_data_inc(incFile, data)

    # Write settings structure
    write_settings_src(srcFile, settings, embedded_flag)
    write_settings_inc(incFile, settings, embedded_flag)

    # Write scaling structure
    write_scaling_src(srcFile, scaling)
    write_scaling_inc(incFile, scaling)

    # Write linsys_solver structure
    write_linsys_solver_src(srcFile, linsys_solver, embedded_flag)
    write_linsys_solver_inc(incFile, linsys_solver, embedded_flag)

    # Define empty solution structure
    write_solution_src(srcFile, data)
    write_solution_inc(incFile, data)

    # Define info structure
    write_info_src(srcFile)
    write_info_inc(incFile)

    # Define workspace structure
    write_workspace_src(srcFile, n, m, rho_vectors, embedded_flag)
    write_workspace_inc(incFile, n, m, rho_vectors, embedded_flag)

    # The endif for the include-guard
    incFile.write("#endif // ifndef %s\n" % incGuard)

    incFile.close()
    srcFile.close()


def render_setuppy(variables, output):
    """
    Render setup.py file
    """

    embedded_flag = variables['embedded_flag']
    python_ext_name = variables['python_ext_name']

    f = open(os.path.join(files_to_generate_path, 'setup.py'))
    filedata = f.read()
    f.close()

    filedata = filedata.replace("EMBEDDED_FLAG", str(embedded_flag))
    filedata = filedata.replace("PYTHON_EXT_NAME", str(python_ext_name))

    f = open(output, 'w')
    f.write(filedata)
    f.close()


def render_cmakelists(variables, output):
    """
    Render CMakeLists file
    """

    embedded_flag = variables['embedded_flag']

    f = open(os.path.join(files_to_generate_path, 'CMakeLists.txt'))
    filedata = f.read()
    f.close()

    filedata = filedata.replace("EMBEDDED_FLAG", str(embedded_flag))

    f = open(output, 'w')
    f.write(filedata)
    f.close()


def render_emosqpmodule(variables, output):
    """
    Render emosqpmodule.c file
    """

    python_ext_name = variables['python_ext_name']

    f = open(os.path.join(files_to_generate_path, 'emosqpmodule.c'))
    filedata = f.read()
    f.close()

    filedata = filedata.replace("PYTHON_EXT_NAME", str(python_ext_name))

    f = open(output, 'w')
    f.write(filedata)
    f.close()



# Test osqp python module
import osqp
from osqp._osqp import constant
# import osqppurepy as osqp
from scipy import sparse
import numpy as np

# Unit Test
import unittest


class primal_infeasibility_tests(unittest.TestCase):

    def setUp(self):
        np.random.seed(6)
        """
        Setup primal infeasible problem
        """

        self.opts = {'verbose': False,
                     'eps_abs': 1e-05,
                     'eps_rel': 1e-05,
                     'eps_dual_inf': 1e-20,
                     'max_iter': 2500,
                     'polish': False}

    def test_primal_infeasible_problem(self):

        # Simple QP problem
        np.random.seed(4)

        self.n = 50
        self.m = 500
        # Generate random Matrices
        Pt = np.random.rand(self.n, self.n)
        self.P = sparse.triu(Pt.T.dot(Pt), format='csc')
        self.q = np.random.rand(self.n)
        self.A = sparse.random(self.m, self.n).tolil()  # Lil for efficiency
        self.u = 3 + np.random.randn(self.m)
        self.l = -3 + np.random.randn(self.m)

        # Make random problem primal infeasible
        self.A[int(self.n/2), :] = self.A[int(self.n/2)+1, :]
        self.l[int(self.n/2)] = self.u[int(self.n/2)+1] + 10 * np.random.rand()
        self.u[int(self.n/2)] = self.l[int(self.n/2)] + 0.5

        # Convert A to csc
        self.A = self.A.tocsc()

        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Solve problem with OSQP
        res = self.model.solve()

        # Assert close
        self.assertEqual(res.info.status_val,
                         constant('OSQP_PRIMAL_INFEASIBLE'))

    def test_primal_and_dual_infeasible_problem(self):

        self.n = 2
        self.m = 4
        self.P = sparse.csc_matrix((2, 2))
        self.q = np.array([-1., -1.])
        self.A = sparse.csc_matrix([[1., -1.], [-1., 1.], [1., 0.], [0., 1.]])
        self.l = np.array([1., 1., 0., 0.])
        self.u = np.inf * np.ones(self.m)

        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        res = self.model.solve()

        # Assert close
        self.assertIn(res.info.status_val, [constant('OSQP_PRIMAL_INFEASIBLE'),
                                            constant('OSQP_DUAL_INFEASIBLE')])



# Test osqp python module
import osqp
from osqp.tests.utils import load_high_accuracy, rel_tol, abs_tol, decimal_tol
# import osqppurepy as osqp
import numpy as np
from scipy import sparse

# Unit Test
import unittest
import numpy.testing as nptest


class polish_tests(unittest.TestCase):

    def setUp(self):
        """
        Setup default options
        """
        self.opts = {'verbose': False,
                     'eps_abs': 1e-03,
                     'eps_rel': 1e-03,
                     'scaling': True,
                     'rho': 0.1,
                     'alpha': 1.6,
                     'max_iter': 2500,
                     'polish': True,
                     'polish_refine_iter': 4}

    def test_polish_simple(self):

        # Simple QP problem
        self.P = sparse.diags([11., 0.], format='csc')
        self.q = np.array([3, 4])
        self.A = sparse.csc_matrix(
            [[-1, 0], [0, -1], [-1, -3], [2, 5], [3, 4]])
        self.u = np.array([0, 0, -15, 100, 80])
        self.l = -1e05 * np.ones(len(self.u))
        self.n = self.P.shape[0]
        self.m = self.A.shape[0]
        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Solve problem
        res = self.model.solve()

        x_sol, y_sol, obj_sol = load_high_accuracy('test_polish_simple')
        # Assert close
        nptest.assert_allclose(res.x, x_sol, rtol=rel_tol, atol=abs_tol)
        nptest.assert_allclose(res.y, y_sol, rtol=rel_tol, atol=abs_tol)
        nptest.assert_almost_equal(
            res.info.obj_val, obj_sol, decimal=decimal_tol)

    def test_polish_unconstrained(self):

        # Unconstrained QP problem
        np.random.seed(4)

        self.n = 30
        self.m = 0
        P = sparse.diags(np.random.rand(self.n)) + 0.2*sparse.eye(self.n)
        self.P = P.tocsc()
        self.q = np.random.randn(self.n)
        self.A = sparse.csc_matrix((self.m, self.n))
        self.l = np.array([])
        self.u = np.array([])
        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Solve problem
        res = self.model.solve()

        x_sol, _, obj_sol = load_high_accuracy('test_polish_unconstrained')
        # Assert close
        nptest.assert_allclose(res.x, x_sol, rtol=rel_tol, atol=abs_tol)
        nptest.assert_almost_equal(
            res.info.obj_val, obj_sol, decimal=decimal_tol)

    def test_polish_random(self):

        # Random QP problem
        np.random.seed(6)

        self.n = 30
        self.m = 50
        Pt = sparse.random(self.n, self.n)
        self.P = Pt.T @ Pt
        self.q = np.random.randn(self.n)
        self.A = sparse.csc_matrix(np.random.randn(self.m, self.n))
        self.l = -3 + np.random.randn(self.m)
        self.u = 3 + np.random.randn(self.m)
        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Solve problem
        res = self.model.solve()

        x_sol, y_sol, obj_sol = load_high_accuracy('test_polish_random')
        # Assert close
        nptest.assert_allclose(res.x, x_sol, rtol=rel_tol, atol=abs_tol)
        nptest.assert_allclose(res.y, y_sol, rtol=rel_tol, atol=abs_tol)
        nptest.assert_almost_equal(
            res.info.obj_val, obj_sol, decimal=decimal_tol)



# from osqp import __path__
from __future__ import print_function
import osqp
import os.path
import shutil as sh
from subprocess import call
from glob import glob
from platform import system
import sys

# import utilities
from . import utils


def codegen(work, target_dir, python_ext_name, project_type, compile_python_ext, embedded,
            force_rewrite, float_flag, long_flag):
    """
    Generate code
    """

    # Import OSQP path
    osqp_path = osqp.__path__[0]

    # Path of osqp module
    files_to_generate_path = os.path.join(osqp_path,
                                          'codegen', 'files_to_generate')

    # Module extension
    if system() == 'Linux' or system() == 'Darwin':
        module_ext = '.so'
    else:
        module_ext = '.pyd'

    # Check if interface already exists
    resp = None   # Initialize response
    if os.path.isdir(target_dir):
        if force_rewrite:
            sh.rmtree(target_dir)
        else:
            while resp != 'n' and resp != 'y':
                resp = input("Directory \"%s\" already exists." %
                             target_dir +
                             " Do you want to replace it? [y/n] ")
                if resp == 'y':
                    sh.rmtree(target_dir)

    # Check if python module already exists
    resp = None  # Initialize response
    if any(glob('%s*%s' % (python_ext_name, module_ext))):
        module_name = glob('%s*%s' % (python_ext_name, module_ext))[0]
        if force_rewrite:
            os.remove(module_name)
        else:
            while resp != 'n' and resp != 'y':
                resp = input("Python module \"%s\" already exists." %
                             module_name +
                             " Do you want to replace it? [y/n] ")
                if resp == 'y':
                    os.remove(module_name)

    # Make target directory
    sys.stdout.write("Creating target directories... \t\t\t\t\t")
    sys.stdout.flush()
    target_dir = os.path.abspath(target_dir)
    target_include_dir = os.path.join(target_dir, 'include')
    target_configure_dir = os.path.join(target_dir, 'configure')
    target_include_dir = os.path.join(target_dir, 'include')
    target_src_dir = os.path.join(target_dir, 'src')

    if not os.path.exists(target_dir):
        os.mkdir(target_dir)
    if not os.path.exists(target_include_dir):
            os.mkdir(target_include_dir)
    if not os.path.exists(target_configure_dir):
            os.mkdir(target_configure_dir)
    if not os.path.exists(target_src_dir):
        os.makedirs(os.path.join(target_src_dir, 'osqp'))
    print("[done]")

    # Copy source files to target directory
    sys.stdout.write("Copying OSQP sources... \t\t\t\t\t")
    sys.stdout.flush()
    c_sources = glob(os.path.join(osqp_path, 'codegen', 'sources',
                                  'src', '*.c'))
    if embedded == 1:
        # Remove kkt.c from embedded sources
        c_sources.remove(os.path.join(osqp_path, 'codegen', 'sources',
                                      'src', 'kkt.c'))
    for source in c_sources:
        sh.copy(source, os.path.join(target_src_dir, 'osqp'))

    # Copy header files
    c_headers = glob(os.path.join(osqp_path, 'codegen', 'sources',
                                  'include', '*.h'))
    if embedded == 1:
        # Remove kkt.h from embedded sources
        c_headers.remove(os.path.join(osqp_path, 'codegen', 'sources',
                                      'include', 'kkt.h'))
    for header in c_headers:
        sh.copy(header, target_include_dir)

    # Copy config files
    c_configs = glob(os.path.join(osqp_path, 'codegen', 'sources',
                                  'configure', '*.h.in'))
    for config in c_configs:
        sh.copy(config, target_configure_dir)

    print("[done]")

    # Variables created from the workspace
    sys.stdout.write("Generating customized code... \t\t\t\t\t")
    sys.stdout.flush()
    template_vars = {'rho_vectors':     work['rho_vectors'],
                     'data':            work['data'],
                     'settings':        work['settings'],
                     'linsys_solver':   work['linsys_solver'],
                     'scaling':         work['scaling'],
                     'embedded_flag':   embedded,
                     'python_ext_name': python_ext_name}

    # Add cmake args
    cmake_args = '-DEMBEDDED:INT=%d -DDFLOAT:BOOL=%s -DDLONG:BOOL=%s' % \
                 (embedded, float_flag, long_flag)

    # Render workspace
    utils.render_workspace(template_vars,
                           os.path.join(target_include_dir, 'workspace.h'),
                           os.path.join(target_src_dir, 'osqp', 'workspace.c'))

    # Render setup.py
    utils.render_setuppy(template_vars,
                         os.path.join(target_src_dir, 'setup.py'))

    # Render emosqpmodule.c
    utils.render_emosqpmodule(template_vars,
                              os.path.join(target_src_dir,
                                           '%smodule.c' % python_ext_name))

    # Render CMakelists.txt
    utils.render_cmakelists(template_vars,
                            os.path.join(target_dir, 'CMakeLists.txt'))

    # Copy cmake files
    sh.copy(os.path.join(osqp_path, 'codegen', 'sources', 'src', 'CMakeLists.txt'),
            os.path.join(target_src_dir, 'osqp'))
    sh.copy(os.path.join(osqp_path, 'codegen', 'sources', 'include', 'CMakeLists.txt'),
            os.path.join(target_include_dir))

    # Copy example.c
    sh.copy(os.path.join(files_to_generate_path, 'example.c'), target_src_dir)

    print("[done]")

    # Create a project
    if project_type != '':
        sys.stdout.write("Creating project...\n")
        sys.stdout.flush()
        current_dir = os.getcwd()
        os.chdir(target_dir)
        if os.path.exists('build'):
            sh.rmtree('build')
        os.makedirs('build')
        os.chdir('build')
        call(['cmake', cmake_args, '-G', project_type, '..'])
        os.chdir(current_dir)
        print("[done]")

    # Compile python interface and copy compiled solver
    if compile_python_ext:
        sys.stdout.write("Compiling Python wrapper... \t\t\t\t\t")
        sys.stdout.flush()
        current_dir = os.getcwd()
        os.chdir(target_src_dir)
        call([sys.executable, 'setup.py', '--quiet', 'build_ext', '--inplace'])
        print("[done]")
        sys.stdout.write("Copying code-generated Python solver to current " +
                         "directory... \t")
        sys.stdout.flush()
        module_name = glob('%s*' % python_ext_name + module_ext)
        if not any(module_name):
            raise ValueError('No Python module generated! ' +
                             'Some errors have occurred.')
        module_name = module_name[0]
        sh.copy(module_name, current_dir)
        os.chdir(current_dir)
        print("[done]")



# Test osqp python module
import osqp
from osqp._osqp import constant
# import osqppurepy as osqp
import numpy as np
from scipy import sparse

# Unit Test
import unittest


class dual_infeasibility_tests(unittest.TestCase):

    def setUp(self):
        """
        Setup default options
        """
        self.opts = {'verbose': False,
                     'eps_abs': 1e-05,
                     'eps_rel': 1e-05,
                     'eps_prim_inf': 1e-15,  # Focus only on dual infeasibility
                     'eps_dual_inf': 1e-6,
                     'scaling': 3,
                     'max_iter': 2500,
                     'polish': False,
                     'check_termination': 1,
                     'polish_refine_iter': 4}

    def test_dual_infeasible_lp(self):

        # Dual infeasible example
        self.P = sparse.csc_matrix((2, 2))
        self.q = np.array([2, -1])
        self.A = sparse.eye(2, format='csc')
        self.l = np.array([0., 0.])
        self.u = np.array([np.inf, np.inf])

        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Solve problem with OSQP
        res = self.model.solve()

        # Assert close
        self.assertEqual(res.info.status_val,
                         constant('OSQP_DUAL_INFEASIBLE'))

    def test_dual_infeasible_qp(self):

        # Dual infeasible example
        self.P = sparse.diags([4., 0.], format='csc')
        self.q = np.array([0, 2])
        self.A = sparse.csc_matrix([[1., 1.], [-1., 1.]])
        self.l = np.array([-np.inf, -np.inf])
        self.u = np.array([2., 3.])

        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Solve problem with OSQP
        res = self.model.solve()

        # Assert close
        self.assertEqual(res.info.status_val,
                         constant('OSQP_DUAL_INFEASIBLE'))

    def test_primal_and_dual_infeasible_problem(self):

        self.n = 2
        self.m = 4
        self.P = sparse.csc_matrix((2, 2))
        self.q = np.array([-1., -1.])
        self.A = sparse.csc_matrix([[1., -1.], [-1., 1.], [1., 0.], [0., 1.]])
        self.l = np.array([1., 1., 0., 0.])
        self.u = np.inf * np.ones(self.m)

        self.model = osqp.OSQP()
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u,
                         **self.opts)

        # Warm start to avoid infeasibility detection at first step
        x0 = 25.*np.ones(self.n)
        y0 = -2.*np.ones(self.m)
        self.model.warm_start(x=x0, y=y0)

        # Solve
        res = self.model.solve()

        # Assert close
        self.assertIn(res.info.status_val, [constant('OSQP_PRIMAL_INFEASIBLE'),
                                            constant('OSQP_DUAL_INFEASIBLE')])



import osqp
import numpy as np
from scipy import sparse

# Unit Test
import unittest


class mkl_pardiso_tests(unittest.TestCase):

    def setUp(self):

        # Simple QP problem
        self.P = sparse.csc_matrix([[3.,  2.],
                                    [2.,  3.]]
                                   )
        self.q = np.array([1.0, 1.0])
        self.A = sparse.csc_matrix([[1.0, 0.0], [0.0, 1.0]])
        self.l = np.array([0.0, 0.0])
        self.u = np.array([100.0, 100.0])

    def test_issue14(self):

        m = osqp.OSQP()
        m.setup(self.P, self.q, self.A, self.l, self.u,
                linsys_solver="mkl pardiso")
        m.solve()

        #  # Assert test_setup flag
        #  self.assertEqual(test_setup, 0)



# Test osqp python module
import osqp
# import osqppurepy as osqp
import numpy as np
import numpy.random as npr
from scipy import sparse
from scipy.optimize import approx_fprime
import numpy.testing as npt

# Unit Test
import unittest


npr.seed(1)

# Tests settings
grad_precision = 1e-5
rel_tol = 1e-3
abs_tol = 1e-3

# OSQP settings
eps_abs = 1e-10
eps_rel = 1e-10
max_iter = 10000


class derivative_tests(unittest.TestCase):

    def get_prob(self, n=10, m=3, P_scale=1., A_scale=1.):
        L = np.random.randn(n, n)
        P = sparse.csc_matrix(L.dot(L.T) + 5. * sparse.eye(n))
        x_0 = npr.randn(n)
        s_0 = npr.rand(m)
        A = sparse.csc_matrix(npr.randn(m, n))
        u = A.dot(x_0) + s_0
        l = -5 - 10 * npr.rand(m)
        q = npr.randn(n)
        true_x = npr.randn(n)

        return [P, q,  A, l, u, true_x]

    def get_grads(self, P, q, A, l, u, true_x):
        # Get gradients by solving with osqp
        m = osqp.OSQP()
        m.setup(P, q, A, l, u, eps_abs=eps_abs, eps_rel=eps_rel,
                max_iter=max_iter, verbose=False)
        results = m.solve()
        if results.info.status != "solved":
            raise ValueError("Problem not solved!")
        x = results.x
        grads = m.adjoint_derivative(dx=x - true_x)

        return grads

    def test_dl_dq(self, verbose=False):
        n, m = 5, 5

        prob = self.get_prob(n=n, m=m, P_scale=100., A_scale=100.)
        P, q, A, l, u, true_x = prob

        def grad(q):
            [dP, dq, dA, dl, du] = self.get_grads(P, q, A, l, u, true_x)
            return dq

        def f(q):
            m = osqp.OSQP()
            m.setup(P, q, A, l, u, eps_abs=eps_abs, eps_rel=eps_rel,
                    max_iter=max_iter, verbose=False)
            res = m.solve()
            if res.info.status != "solved":
                raise ValueError("Problem not solved!")
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dq = grad(q)
        dq_fd = approx_fprime(q, f, grad_precision)

        if verbose:
            print('dq_fd: ', np.round(dq_fd, decimals=4))
            print('dq: ', np.round(dq, decimals=4))

        npt.assert_allclose(dq_fd, dq, rtol=rel_tol, atol=abs_tol)

    def test_dl_dP(self, verbose=False):
        n, m = 3, 3

        prob = self.get_prob(n=n, m=m, P_scale=100., A_scale=100.)
        P, q, A, l, u, true_x = prob
        P_idx = P.nonzero()

        def grad(P_val):
            P_qp = sparse.csc_matrix((P_val, P_idx), shape=P.shape)
            [dP, dq, dA, dl, du] = self.get_grads(P_qp, q, A, l, u, true_x)
            return dP

        def f(P_val):
            P_qp = sparse.csc_matrix((P_val, P_idx), shape=P.shape)
            m = osqp.OSQP()
            m.setup(P_qp, q, A, l, u, eps_abs=eps_abs,
                    eps_rel=eps_rel, max_iter=max_iter, verbose=False)
            res = m.solve()
            if res.info.status != "solved":
                raise ValueError("Problem not solved!")
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dP = grad(P.data)
        dP_fd_val = approx_fprime(P.data, f, grad_precision)
        dP_fd = sparse.csc_matrix((dP_fd_val, P_idx), shape=P.shape)
        dP_fd = (dP_fd + dP_fd.T)/2

        if verbose:
            print('dP_fd: ', np.round(dP_fd.data, decimals=4))
            print('dA: ', np.round(dP.data, decimals=4))

        npt.assert_allclose(dP.todense(), dP_fd.todense(),
                            rtol=rel_tol, atol=abs_tol)

    def test_dl_dA(self, verbose=False):
        n, m = 3, 3

        prob = self.get_prob(n=n, m=m, P_scale=100., A_scale=100.)
        P, q, A, l, u, true_x = prob
        A_idx = A.nonzero()

        def grad(A_val):
            A_qp = sparse.csc_matrix((A_val, A_idx), shape=A.shape)
            [dP, dq, dA, dl, du] = self.get_grads(P, q, A_qp, l, u, true_x)
            return dA

        def f(A_val):
            A_qp = sparse.csc_matrix((A_val, A_idx), shape=A.shape)
            m = osqp.OSQP()
            m.setup(P, q, A_qp, l, u, eps_abs=eps_abs,
                    eps_rel=eps_rel, max_iter=max_iter, verbose=False)
            res = m.solve()
            if res.info.status != "solved":
                raise ValueError("Problem not solved!")
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dA = grad(A.data)
        dA_fd_val = approx_fprime(A.data, f, grad_precision)
        dA_fd = sparse.csc_matrix((dA_fd_val, A_idx), shape=A.shape)

        if verbose:
            print('dA_fd: ', np.round(dA_fd.data, decimals=4))
            print('dA: ', np.round(dA.data, decimals=4))

        npt.assert_allclose(dA.todense(), dA_fd.todense(),
                            rtol=rel_tol, atol=abs_tol)

    def test_dl_dl(self, verbose=False):
        n, m = 10, 10

        prob = self.get_prob(n=n, m=m, P_scale=100., A_scale=100.)
        P, q, A, l, u, true_x = prob

        def grad(l):
            [dP, dq, dA, dl, du] = self.get_grads(P, q, A, l, u, true_x)
            return dl

        def f(l):
            m = osqp.OSQP()
            m.setup(P, q, A, l, u, eps_abs=eps_abs, eps_rel=eps_rel,
                    max_iter=max_iter, verbose=False)
            res = m.solve()
            if res.info.status != "solved":
                raise ValueError("Problem not solved!")
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dl = grad(l)
        dl_fd = approx_fprime(l, f, grad_precision)

        if verbose:
            print('dl_fd: ', np.round(dl_fd, decimals=4))
            print('dl: ', np.round(dl, decimals=4))

        npt.assert_allclose(dl_fd, dl,
                            rtol=rel_tol, atol=abs_tol)

    def test_dl_du(self, verbose=False):
        n, m = 5, 5

        prob = self.get_prob(n=n, m=m, P_scale=100., A_scale=100.)
        P, q, A, l, u, true_x = prob

        def grad(u):
            [dP, dq, dA, dl, du] = self.get_grads(P, q, A, l, u, true_x)
            return du

        def f(u):
            m = osqp.OSQP()
            m.setup(P, q, A, l, u, eps_abs=eps_abs, eps_rel=eps_rel,
                    max_iter=max_iter, verbose=False)
            res = m.solve()
            if res.info.status != "solved":
                raise ValueError("Problem not solved!")
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        du = grad(u)
        du_fd = approx_fprime(u, f, grad_precision)

        if verbose:
            print('du_fd: ', np.round(du_fd, decimals=4))
            print('du: ', np.round(du, decimals=4))

        npt.assert_allclose(du_fd, du,
                            rtol=rel_tol, atol=abs_tol)



"""Common utility functions"""
from warnings import warn
import numpy as np
import scipy.sparse as sparse
import osqp._osqp as _osqp


def linsys_solver_str_to_int(settings):
        linsys_solver_str = settings.pop('linsys_solver', '')
        if not isinstance(linsys_solver_str, str):
            raise TypeError("Setting linsys_solver " +
                            "is required to be a string.")
        linsys_solver_str = linsys_solver_str.lower()
        if linsys_solver_str == 'qdldl':
            settings['linsys_solver'] = _osqp.constant('QDLDL_SOLVER')
        elif linsys_solver_str == 'mkl pardiso':
            settings['linsys_solver'] = _osqp.constant('MKL_PARDISO_SOLVER')
        # Default solver: QDLDL
        elif linsys_solver_str == '':
            settings['linsys_solver'] = _osqp.constant('QDLDL_SOLVER')
        else:   # default solver: QDLDL
            warn("Linear system solver not recognized. " +
                 "Using default solver QDLDL.")
            settings['linsys_solver'] = _osqp.constant('QDLDL_SOLVER')
        return settings


def prepare_data(P=None, q=None, A=None, l=None, u=None, **settings):
        """
        Prepare problem data of the form

        minimize     1/2 x' * P * x + q' * x
        subject to   l <= A * x <= u

        solver settings can be specified as additional keyword arguments
        """

        #
        # Get problem dimensions
        #

        if P is None:
            if q is not None:
                n = len(q)
            elif A is not None:
                n = A.shape[1]
            else:
                raise ValueError("The problem does not have any variables")
        else:
            n = P.shape[0]
        if A is None:
            m = 0
        else:
            m = A.shape[0]

        #
        # Create parameters if they are None
        #

        if (A is None and (l is not None or u is not None)) or \
                (A is not None and (l is None and u is None)):
            raise ValueError("A must be supplied together " +
                             "with at least one bound l or u")

        # Add infinity bounds in case they are not specified
        if A is not None and l is None:
            l = -np.inf * np.ones(A.shape[0])
        if A is not None and u is None:
            u = np.inf * np.ones(A.shape[0])

        # Create elements if they are not specified
        if P is None:
            P = sparse.csc_matrix((np.zeros((0,), dtype=np.double),
                                   np.zeros((0,), dtype=int),
                                   np.zeros((n+1,), dtype=int)),
                                  shape=(n, n))
        if q is None:
            q = np.zeros(n)

        if A is None:
            A = sparse.csc_matrix((np.zeros((0,), dtype=np.double),
                                   np.zeros((0,), dtype=int),
                                   np.zeros((n+1,), dtype=int)),
                                  shape=(m, n))
            l = np.zeros(A.shape[0])
            u = np.zeros(A.shape[0])

        #
        # Check vector dimensions (not checked from C solver)
        #

        # Check if second dimension of A is correct
        # if A.shape[1] != n:
        #     raise ValueError("Dimension n in A and P does not match")
        if len(q) != n:
            raise ValueError("Incorrect dimension of q")
        if len(l) != m:
            raise ValueError("Incorrect dimension of l")
        if len(u) != m:
            raise ValueError("Incorrect dimension of u")

        #
        # Check or Sparsify Matrices
        #
        if not sparse.issparse(P) and isinstance(P, np.ndarray) and \
                len(P.shape) == 2:
            raise TypeError("P is required to be a sparse matrix")
        if not sparse.issparse(A) and isinstance(A, np.ndarray) and \
                len(A.shape) == 2:
            raise TypeError("A is required to be a sparse matrix")

        # If P is not triu, then convert it to triu
        if sparse.tril(P, -1).data.size > 0:
            P = sparse.triu(P, format='csc')

        # Convert matrices in CSC form and to individual pointers
        if not sparse.isspmatrix_csc(P):
            warn("Converting sparse P to a CSC " +
                 "(compressed sparse column) matrix. (It may take a while...)")
            P = P.tocsc()
        if not sparse.isspmatrix_csc(A):
            warn("Converting sparse A to a CSC " +
                 "(compressed sparse column) matrix. (It may take a while...)")
            A = A.tocsc()

        # Check if P an A have sorted indices
        if not P.has_sorted_indices:
            P.sort_indices()
        if not A.has_sorted_indices:
            A.sort_indices()

        # Convert infinity values to OSQP Infinity
        u = np.minimum(u, _osqp.constant('OSQP_INFTY'))
        l = np.maximum(l, -_osqp.constant('OSQP_INFTY'))

        # Convert linsys_solver string to integer
        settings = linsys_solver_str_to_int(settings)

        return ((n, m), P.data, P.indices, P.indptr, q,
                A.data, A.indices, A.indptr,
                l, u), settings
